############################################################
# IQUASAP-CCES
# Functions to process and plot permafrost borehole temperature data
#
# - filter.min()
# - filter.max()
# - filter.minmax()
# - eval.zero.curtain()
# - corr.zero.curtain()
# 
# Author: Rachel Lthi
# Created: 24.10.2014
# Modified: 17.01.2016
#
############################################################

# Load required packages
require("caTools")
require("zoo")

#-----------------------------------------------------------
# function
#
# filter.min<-function(data,min,delete=T,repval=min)
# 
# Replaces data : x<=min <- NA if delete=T or x<=min <- repval if delete = F
#
# Variables:
# - data: data matrix (or vector), numeric
# - min: minimum value, numeric
# - delete: wheter to delete or replace value, boolean
# - repval: replacement value, numeric
#
# Output:
# - data matrix (or vector), numeric; same as input
#-----------------------------------------------------------
filter.min<-function(data,min,delete=T,repval=min){
  if(delete){
    replace(data,which(data<=min),NA)
  }else{
    replace(data,which(data<=min),repval)
  }
}

#-----------------------------------------------------------
# function
#
# filter.max<-function(data,max,delete=T,repval=max)
# 
# Replaces data : x>=max <- NA if delete=T or x>=max <- repval if delete = F
#
# Variables:
# - data: data matrix (or vector), numeric
# - max: maximum value, numeric
# - delete: wheter to delete or replace value, boolean
# - repval: replacement value, numeric
#
# Output:
# - data matrix (or vector), numeric; same as input
#-----------------------------------------------------------
filter.max<-function(data,max,delete=T,repval=max){
  if(delete){
    replace(data,which(data>=max),NA)
  }else{
    replace(data,which(data>=max),repval)
  }
}

#-----------------------------------------------------------
# function
#
# filter.minmax<-function(data,min,max,delete=T,repval=c(min,max))
# 
# Replaces data with NA (if delete = T): (x<=min || x>=max) <- NA
# or with the boarder values if delete = F: x<=min <- repval[1] & x>=max <- repval[2]
#
# Variables:
# - data: data matrix (or vector), numeric
# - min: minimum value, numeric 
# - max: maximum value, numeric
# - delete: wheter to delete or replace value, boolean
# - repval: (min, max); values to replace data bigger than the specified max value, or smaller than the min value
#
# Output:
# - data matrix (or vector), numeric; same as input
#-----------------------------------------------------------
filter.minmax<-function(data,min,max,delete=T,repval=c(min,max)){
  if(delete){
    data<-replace(data,which(data<=min),NA)
    data<-replace(data,which(data>=max),NA)
  }else{
    data<-replace(data,which(data<=min),repval[1])
    data<-replace(data,which(data>=max),repval[2])
  }
  data
}

#-----------------------------------------------------------
# function
#
# eval.zero.curtain<-function(date,data,prec=0.01,range=c(-0.5,0.5),k=10,plot=TRUE,dateFormat="%Y-%m-%d",site="")
# 
# Evaluates time windows with zero curtain
#
# Variables:
# - date: date vector
# - data: data vector
# - prec: precision for zero curtain, numeric
# - range: range for zero curtain, numeric vector
# - k: runmean window
# - plot: TRUE/FALSE
# - dateFormat: default = "%Y-%m-%d"; specify the format of the date vector
# - site: Plot title if plotting is enabled
# 
# Output:
# - out: matrix containing date, is zero curtain (boolean), input data, running mean values
#-----------------------------------------------------------

eval.zero.curtain<-function(date,data,prec=0.01,range=c(-0.5,0.5),k=10,plot=TRUE,dateFormat="%Y-%m-%d",site=""){
  values<-runmean(data,k,alg="exact",endrule="mean") # Calculations are done using running mean
  if(sum(values>0,na.rm=TRUE)>0 && sum(values<0,na.rm=TRUE)>0){ # Condition 1: Values >0 and <0
    l<-length(values)-1
    for(i in 1:l){ # test prec (precision for zero curtain)
      if(is.na(values[i])||is.na(values[i+1])||(is.na(values[i]&&is.na(values[i+1])))){
        x<-NA
      }else{
        if(abs(values[i]-values[i+1])<=prec){x<-1}else{x<-0}
      }
      if(i==1){prec<-x}else{prec<-c(prec,x)}
    } # end test prec
    
    # test range (range for zero curtain)
    tr<-values
    tr<-replace(tr,which(tr<=range[2] & tr>=(range[1])),1)
    tr<-replace(tr,which(tr!=1 && !is.na(tr)),0)
    tr<-tr[1:l]
    
    # format date
    date<-as.Date(date[2:length(date)],dateFormat)
    mt<-as.numeric(substr(as.character(date),6,7))
    
    print(paste("Start date: ",date[1],", End date: ",date[length(date)]),sep="")
    print("Months:")
    print(head(mt))
    print((paste("Is.numeric:",is.numeric(mt),sep="")))
    
    # define months used for zero curtain detection
    activemonths<-replace(mt,mt %in% c(1,2,3,9,12),0) #Inactive months (not used for zero curtain detection)
    activemonths<-replace(activemonths,activemonths %in% c(4,5,6,7,8,10,11),1) #Active months (used for zero curtain detection)
    
    print(head(activemonths))
    
    # combine all criteria for the zero curtain detection (1: zero curtain, 0: not zero curtain)
    iszc<--999
    for (i in 1:l){
      if(is.na(tr[i])){
        iszc[i]<-NA
      }else{
        if((prec[i]==1 && tr[i]==1 && activemonths[i]==1)||(is.na(prec[i]) && tr[i]==1 && activemonths[i]==1)){
          iszc[i]<-1
        }else{
          iszc[i]<-0
        }
      }
    }
    
    # assign values to detected zero curtains
    tplot<-iszc
    for(i in 1:l){
      if(!is.na(iszc[i])){
        if(iszc[i]==0){tplot[i]<-NA}
        if(iszc[i]==1){tplot[i]<-values[i]}
      } 
    }
    
    # plot zero curtain detection
    if(plot){
      plot(date,data[2:length(data)],main=site,xaxt="naxt",xlab="Time",ylab="Temperature (C)",col="grey",type="l")
      abline(h=0)
      points(date,values[1:l],type="l")
      points(date,tplot,col="red",type="l",lwd=2)
      legend("topleft",c("measured data","running mean of measured data","zero curtain"),col=c("grey","black","red"),lwd=1,lty=1,bty="n")
    } # end plot
  } else {iszc<-rep(-1,length(values)-1)}
  
  # define output matrix
  out<-cbind(date,iszc,data[2:length(data)],values[1:l])
  # return output matrix
  out
} # end function

#-----------------------------------------------------------
# function
#
# corr.zero.curtain<-function(data,model="linear",plot=TRUE,site="")
# 
# Correction based on the detected zero curtains using a specified model
#
# Variables:
# - data: data matrix, numeric (Output given by eval.zero.curtain)
# - model: "linear","poly2", "poly3", "polyspline2", "polyspline3", "polyspline4"
# - plot: TRUE/FALSE
# - site: site name for plot title
# 
# Output:
# out: matrix containing date, corrected data, flag if corresponding data has been corrected
# 
#-----------------------------------------------------------

corr.zero.curtain<-function(data,model="linear",plot=TRUE,site=""){
  date<-data[,1]
  zc<-data[,2]
  bht<-data[,3]
  values<-replace(bht,zc!=1,NA)
  x<-seq(1,length(values),by=1)
  
  # approximation function (different models)
  
  if(model=="linear"){
    fm<-lm(values~x,na.action = na.exclude)
    out_approx<-na.approx(fitted.values(fm),rule=2)
  }
  
  if(model=="poly2"){
    fm<-lm(values ~ poly(x,2),na.action = na.exclude)
    out_approx<-na.approx(fitted.values(fm),rule=2)
  }
  
  if(model=="polyspline2"){
    fm<-lm(values ~ poly(x,2),na.action = na.exclude)
    out_approx<-na.spline(fitted.values(fm),rule=2)
  }
  
  if(model=="poly3"){
    fm<-lm(values ~ poly(x,3),na.action = na.exclude)
    out_approx<-na.approx(fitted.values(fm),rule=2)
  }
  
  if(model=="polyspline3"){
    fm<-lm(values ~ poly(x,3),na.action = na.exclude)
    out_approx<-na.spline(fitted.values(fm),rule=2)
  }
  
  if(model=="polyspline4"){
    fm<-lm(values ~ poly(x,4),na.action = na.exclude)
    out_approx<-na.spline(fitted.values(fm),rule=2)
  }
  
  # calculate drift from approximated values
  posdiv<-0
  for(i in 2:length(out_approx)){
    posdiv[i]<-out_approx[i]-out_approx[i-1]
  }
  
  # determine driftbegin
  driftbeg<-min(which(posdiv>0))
  lastNonDrift<-max(which(posdiv<0))
  if(lastNonDrift==-Inf){lastNonDrift<-0}
  
  # assign corrected borehole temperatures and setting flags
  
  if(driftbeg>lastNonDrift){
    out_corrected<-bht
    out_flag<-rep(0,length(bht))
    for (i in driftbeg:length(bht)){
      out_corrected[i]<-bht[i]-(out_approx[i]-out_approx[driftbeg])
      out_flag[i]<-1
    }
  }
  
  if(driftbeg<lastNonDrift){
    out_corrected<-bht
    out_flag<-rep(0,length(bht))
    for (i in lastNonDrift:length(bht)){
      out_corrected[i]<-bht[i]-(out_approx[i]-out_approx[lastNonDrift])
      out_flag[i]<-1
    }
  }
  
  # define output matrix
  out<-cbind(date,out_corrected,out_flag)
  
  # plot results
  if (plot==T) {
    plot(x,values,main=site,xlab="Time",xaxt="naxt",ylab="Ground temperature [C]")
    points(x,(out_approx),type="l",col="red")
    legend("topleft",c("zero curtain","fitted values"),col=c("black","red"),lwd=1,lty=1,bty="n")
  
    plot(x,bht,type="l",col="grey",xaxt="naxt",xlab="Time",ylab="Ground temperature [C]",main=site)
    abline(h=0)
    points(x,out_corrected,type="l")
    legend("topleft",c("measured data","zero curtain corrected data"),col=c("grey","black"),lwd=1,lty=1,bty="n")
  }
  
  # print output summary
  print(c(coef(summary(fm))[2,1],coef(summary(fm))[2,2])) # a,b --> a+b*x
  print(coef(summary(fm))[2,4])
  
  # return output
  out
}# end function